(* infotheo (c) AIST. R. Affeldt, M. Hagiwara, J. Senizergues. GNU GPLv3. *)
From mathcomp Require Import ssreflect ssrbool ssrfun eqtype ssrnat seq div fintype finfun finset.
From mathcomp Require Import bigop prime fingroup zmodp ssralg perm matrix tuple poly.
From mathcomp Require Import finalg mxalgebra mxpoly mxrepresentation binomial.
Require Import ssr_ext ssralg_ext f2 num_occ natbin.

Import GRing.Theory.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Local Open Scope num_occ_scope.

Lemma addb_nseq b : forall r v, size v = r ->
  [seq x.1 (+) x.2 | x <- zip (nseq r b) v] = map (pred1 (negb b)) v.
Proof.
elim; first by case.
move=> r IH [|h t] //= [] r_t.
rewrite IH //; by destruct b; destruct h.
Qed.

Definition addb_seq a b := [seq x.1 (+) x.2 |  x <- zip a b].

Lemma addb_seq_com : forall n a b, size a = n -> size b = n ->
  addb_seq a b = addb_seq b a.
Proof.
elim => [ [] // [] // | n IH [|ha ta] // [|hb tb] // ] [Ha] [Hb].
by rewrite /addb_seq /= -!/(addb_seq _ _) IH // addbC.
Qed.

Lemma addb_tri_ine : forall a b c, a (+) b <= (a (+) c) + (c (+) b).
Proof. by move=> [] [] []. Qed.

Lemma addb_seq_cat : forall a b c d, size a = size c ->
  addb_seq (a ++ b) (c ++ d) = addb_seq a c ++ addb_seq b d.
Proof. move=> a b c d a_c; by rewrite /addb_seq /= -map_cat zip_cat. Qed.

Lemma addb_seq_map : forall n {A : Type} (a b : seq A) f, size a = n ->
  size b = n ->
  addb_seq (map f a) (map f b) = map (fun x => f x.1 (+) f x.2) (zip a b).
Proof.
elim.
  by move=> ? [] // [].
move=> n IH A [|ha ta] // [|hb tb] //= f [Ha] [Hb].
rewrite /addb_seq /=.
f_equal.
by apply IH.
Qed.

Module HammingMetricBitstring.

Definition wH_b (a : bitseq) := N(true | a).

Definition dH_b (a b : bitseq) : nat := wH_b (addb_seq a b).

Lemma dH_b_count : forall n a b, size a = n ->
  dH_b (nseq n b) a = count (pred1 (negb b)) a.
Proof.
elim; first by case.
move=> n IH [|[|] ta] [] // [H]; rewrite /dH_b in IH;
  by rewrite /= /dH_b /= IH.
Qed.

Lemma dH_b_sym : forall n a b, size a = n -> size b = n ->
  dH_b a b = dH_b b a.
Proof. move=> n a b Ha Hb. by rewrite /dH_b (@addb_seq_com n). Qed.

Lemma dH_b_tri_ine : forall n a b c, size a = n -> size b = n -> size c = n ->
  dH_b a b <= dH_b a c + dH_b c b.
Proof.
elim.
  by case=> // [] // [].
move=> n IH [|ha ta] // [|hb tb] // [|hc tc] // [Ha] [Hb] [Hc].
rewrite /dH_b /= -/(dH_b ta tb) -/(dH_b ta tc) -/(dH_b tc tb).
move: {IH}(IH _ _ _ Ha Hb Hc) => IH.
eapply leq_trans.
- apply leq_add; [by apply leqnn | by apply IH].
- rewrite !eqb_id.
  rewrite !addnA -(addnC (hc (+) hb)) addnA -2!addnA; apply leq_add.
  + rewrite addnC; by apply addb_tri_ine.
  + by apply leqnn.
Qed.

Lemma dH_b_cat a b c d : size a = size c -> size b = size d ->
  dH_b (a ++ b) (c ++ d) = dH_b a c + dH_b b d.
Proof. move=> ac bd; by rewrite /dH_b /wH_b addb_seq_cat // /num_occ count_cat. Qed.

End HammingMetricBitstring.

Local Open Scope ring_scope.
Local Open Scope vec_ext_scope.

(* TODO: move *)
Lemma enum_inord (m : nat) : enum 'I_m.+1 = [seq inord i | i <- iota 0 m.+1].
Proof.
rewrite -val_enum_ord -map_comp.
transitivity ([seq i | i <- enum 'I_m.+1]); first by rewrite map_id.
apply eq_map => i /=; by rewrite inord_val.
Qed.

(* TODO: move *)
Lemma seq_of_poly_rV (F : ringType) m (p : {poly F}) (_ : size p < m) :
  [seq (poly_rV p) ord0 i | i <- enum 'I_m] =
  [seq (poly_rV p) ord0 i | i <- enum ('I_(size p))] ++ nseq (m - size p) 0.
Proof.
case: m H => // m pm.
rewrite (_ : enum 'I_m.+1 = map inord (iota 0 (size p) ++ iota (size p) (m.+1 - size p))); last first.
  rewrite -iota_add subnKC; last by rewrite ltnW.
  by rewrite enum_inord.
rewrite 2!map_cat; congr (_ ++ _).
  rewrite -val_enum_ord -!map_comp.
  apply eq_map => i /=.
  rewrite !mxE inordK //.
  by rewrite (ltn_trans (ltn_ord i)).
rewrite -map_comp.
apply (@eq_from_nth _ 0).
  by rewrite size_map size_iota size_nseq.
move=> i.
rewrite size_map size_iota => Hi.
rewrite (nth_map O) ?size_iota //= mxE nth_nseq Hi.
rewrite nth_iota //.
move: (leq_addr i (size p)).
move/leq_sizeP; apply.
rewrite inordK //.
by rewrite -ltn_subRL.
Qed.

Section hamming_weight_distance.

Variables (F : ringType) (n : nat).
Implicit Types u v : 'rV[F]_n.

(* Hamming weight *)
Definition wH v := count (fun x => x != 0) (tuple_of_row v).

Lemma wH_rVpoly u : wH u = count (fun i : 'I_n => (rVpoly u)`_i != 0) (enum 'I_n).
Proof.
rewrite /wH count_map; apply eq_in_count => /= i _.
rewrite coef_poly insubT // => ni.
rewrite ltn_ord; congr (u _ _ != _); by apply val_inj.
Qed.

Lemma wH_poly_rV (p : {poly F}) : wH (poly_rV p) <= size p.
Proof.
rewrite /wH /=.
case/boolP : (size p < n) => pn; last first.
  rewrite -leqNgt in pn.
  by rewrite (leq_trans _ pn) // (leq_trans (count_size _ _)) // size_map size_enum_ord.
have -> : [seq (poly_rV p) ord0 i | i <- enum 'I_n] =
          [seq (poly_rV p) ord0 i | i <- enum ('I_(size p))] ++ nseq (n - size p) 0.
  by rewrite seq_of_poly_rV.
rewrite count_cat.
rewrite [X in _ + X <= _](_ : count _ _ = O) ?addn0; last first.
  rewrite (@eq_in_count _ _ pred0); last first.
    move=> i; case/nseqP => -> /= _; by rewrite eqxx.
  by rewrite count_pred0.
by rewrite count_map (leq_trans (count_size _ _)) // size_enum_ord.
Qed.

Lemma max_wH u : wH u <= n.
Proof. by rewrite /wH (leq_trans (count_size _ _)) // size_tuple. Qed.

Lemma max_wH' u : wH u < n.+1.
Proof. rewrite ltnS. by apply max_wH. Qed.

Lemma wH0 : wH 0 = O.
Proof.
rewrite /wH (@eq_in_count _ _ pred0) ?count_pred0 // => i /mapP[/= j _].
by rewrite mxE => ->; rewrite eqxx.
Qed.

Lemma wH0_inv v : wH v = O -> v = 0.
Proof.
move=> H.
apply/rowP => i; rewrite mxE; apply/eqP; move/eqP : H; apply: contraTT => H.
rewrite -lt0n -has_count; apply/hasP.
exists (v 0 i) => //; apply/mapP; exists i => //.
by rewrite val_ord_tuple mem_enum.
Qed.

Lemma wH_eq0 u : (wH u == O) = (u == 0).
Proof. apply/idP/idP => [/eqP/wH0_inv/eqP //|/eqP ->]; by rewrite wH0. Qed.

Lemma wH_opp v : wH (- v) = wH v.
Proof.
rewrite {1}/wH [X in tval X](_ : _ = [tuple of map (fun x => - x) (tuple_of_row v)]); last first.
  apply: eq_from_tnth => i /=.
  by rewrite !tnth_map mxE.
rewrite count_map; apply eq_count => i /=; by rewrite oppr_eq0.
Qed.

(* Hamming distance *)
Definition dH u v := wH (u - v).

Lemma dH_wH u v : dH u (u + v) = wH v.
Proof. by rewrite /dH opprD addrA subrr add0r wH_opp. Qed.

Lemma dH0 u v : (dH u v == O) = (u == v).
Proof. by rewrite /dH wH_eq0 subr_eq0. Qed.

Lemma max_dH u v : dH u v <= n.
Proof. rewrite /dH. apply max_wH. Qed.

Lemma dH_sym u v : dH u v = dH v u.
Proof. by rewrite {1}/dH -wH_opp opprB. Qed.

Local Open Scope nat_scope.

Lemma wH_sum v : wH v = \sum_(n0 < n) (v ord0 n0 != 0)%R.
Proof.
rewrite /wH 1!count_map -sum1_count /= big_mkcond /=.
apply congr_big => //=; by rewrite /index_enum -enumT.
Qed.

Lemma wH_card_supp u : wH u = #|supp u|.
Proof.
rewrite wH_sum /supp -sum1dep_card [in RHS]big_mkcond; by apply/eq_bigr.
Qed.

Local Close Scope nat_scope.

End hamming_weight_distance.

Section hamming_triangular_inequality.

Variables (F : comRingType).

Lemma tri_ine (a b c : F) : (a != b) <= (c != b) + (a != c).
Proof.
case/boolP : (a == b) => ab //=.
by case/boolP : (c == b) => cb //=; rewrite add0n (eqP cb) {cb} ab.
Qed.

Lemma dH_tri_ine n (c a b : 'rV[F]_n) : (@dH _ n a b <= dH a c + dH c b)%nat.
Proof.
elim: n a b c => [a b c|n IH a b c].
  by rewrite (empty_rV a) (empty_rV b) (empty_rV c) /dH subrr wH0.
rewrite /dH /wH.
have H : forall v w : 'rV[F]_n.+1,
  count (fun x : F => x != 0) (tuple_of_row (v - w)) =
  addn (v 0 0 - w 0 0 != 0) (count (fun x : F => x != 0) (behead (tuple_of_row (v - w)))).
  move=> v w.
  rewrite (tuple_eta (tuple_of_row (v - w))) /=.
  by rewrite /tuple_of_row /thead tnth_map !mxE tnth_ord_tuple.
rewrite !H {H} addnCA !addnA -addnA leq_add //; last first.
move: {IH}(IH (row_of_tuple [tuple of (behead (tuple_of_row a))])
              (row_of_tuple [tuple of (behead (tuple_of_row b))])
              (row_of_tuple [tuple of (behead (tuple_of_row c))])).
  rewrite /dH /wH.
  set lhs := (X in X <= _ -> _).
  set rhs := (X in _ <= X -> _) => IH.
  apply (@leq_trans lhs).
    rewrite /lhs.
    apply eq_leq.
    rewrite (_ : tuple_of_row _ = [tuple of (behead (tuple_of_row (a - b)))]) //.
    apply eq_from_tnth.
    rewrite /tuple_of_row => i; by rewrite !(tnth_behead,tnth_mktuple,mxE).
  apply (leq_trans IH) => {IH}.
  rewrite leq_add //.
  - apply eq_leq.
    rewrite (_ : tuple_of_row _ = [tuple of (behead (tuple_of_row (a - c)))]) //.
    apply eq_from_tnth.
    rewrite /tuple_of_row => i; by rewrite !(tnth_behead,tnth_mktuple,mxE).
  - apply eq_leq.
    rewrite (_ : tuple_of_row _ = [tuple of (behead (tuple_of_row (c - b)))]) //.
    apply eq_from_tnth.
    rewrite /tuple_of_row => i; by rewrite !(tnth_behead,tnth_mktuple,mxE).
by rewrite !subr_eq0 tri_ine.
Qed.

End hamming_triangular_inequality.

Section wH_supp.

Variable n : nat.

Definition wH_supp (x : 'rV['F_2]_n) :=
  mask (map f2.bool_of_F2 (tuple_of_row x)) (enum 'I_n).

Lemma size_wH_supp (x : 'rV['F_2]_n) : size (wH_supp x) = wH x.
Proof.
rewrite /wH_supp size_mask; last first.
  by rewrite size_enum_ord size_map size_tuple.
rewrite count_map -sum1_count wH_sum [RHS](bigID (fun i => x ``_ i == 1%R)) /=.
rewrite [in RHS](eq_bigr (fun=> 1%N)) /=; last by move=> i /eqP -> //.
rewrite [in X in _ = (_ + X)%N](eq_bigr (fun=> 0%N)) /=; last first.
  move=> i; by rewrite -F2_eq0 => /eqP ->.
rewrite [in X in _ = (_ + X)%N]big_const !iter_addn_0 mul0n addn0.
rewrite big_map /index_enum -enumT /=.
apply/eq_bigl => i; by rewrite F2_eq1.
Qed.

Lemma mem_wH_supp (x : 'rV['F_2]_n) (i : 'I_n) : x ``_ i != 0 -> i \in wH_supp x.
Proof.
move=> H'.
rewrite /wH_supp -(cat_take_drop i (enum 'I_n)) -(cat_take_drop i (map _ _)).
rewrite mask_cat; last by rewrite !size_take size_map !size_tuple card_ord.
rewrite mem_cat; apply/orP; right.
rewrite (drop_nth i); last by rewrite size_enum_ord.
rewrite (drop_nth true); last by rewrite size_map /tuple_of_row size_tuple.
rewrite mask_cons mem_cat; apply/orP; left; rewrite nth_ord_enum.
apply/nseqP; split => //.
rewrite lt0n (nth_map 1) ?size_tuple //.
rewrite (_ : nth _ _ _ = x ``_ i); last by rewrite /tuple_of_row nth_mktuple.
apply: contra H'; by case/f2.F2P : (x ``_ i).
Qed.

Lemma wH_suppE x : wH_supp x = mask [seq x ``_ i == 1 | i <- enum 'I_n] (enum 'I_n).
Proof.
rewrite /wH_supp (_ : map _ _ = map (fun i => x``_ i == 1) (enum 'I_n)) //.
rewrite /tuple_of_row -map_comp.
apply eq_map => i; by rewrite /= F2_eq1.
Qed.

End wH_supp.

Section HammingMetric_F_2.

Definition wH_old n (v : 'rV['F_2]_n) := N(1 | tuple_of_row v).

Lemma wH_oldE n (v : 'rV['F_2]_n) : wH v = wH_old v.
Proof.
rewrite /wH /wH_old /num_occ; apply eq_count => i; by rewrite -F2_eq1.
Qed.

Lemma wH_count p (x : 'rV['F_2]_p) : wH x = count (fun i => x ``_ i == 1) (enum 'I_p).
Proof.
rewrite wH_sum -sum1_count [in RHS]big_mkcond /=.
apply congr_big => //.
- by rewrite /index_enum -enumT.
- move=> i _; case: ifPn => [/eqP -> //|].
  by rewrite -F2_eq0 => /eqP ->.
Qed.

Lemma wH_old_wH_b n (a : 'rV_n) :
  wH_old a = HammingMetricBitstring.wH_b (tval (rowF2_tuplebool a)).
Proof.
rewrite /wH_old /HammingMetricBitstring.wH_b /= /num_occ.num_occ /= [in RHS]count_map.
apply eq_in_count => /= x Hx; by rewrite -(F2_eq1 x) eqb_id.
Qed.

Local Open Scope nat_scope.

Lemma sum_wH_row (F : finFieldType) n m (H : 'M[F]_(m, n)) :
  \sum_(m0 : 'I_m) wH (row m0 H) = \sum_(n0 : 'I_n) wH (col n0 H)^T.
Proof.
transitivity (\sum_(m0 < m) \sum_(n0 < n) (H m0 n0 != 0)%R).
  apply eq_bigr => m0 _.
  rewrite wH_sum; apply eq_bigr => n0 _; by rewrite mxE.
rewrite exchange_big /=.
apply eq_bigr => n0 _.
rewrite wH_sum; apply eq_bigr => m0 _; by rewrite 2!mxE.
Qed.

Lemma wH_col_1 n (i : 'I_n) : @wH [fieldType of 'F_2] _ (col i 1%:M)^T = 1.
Proof.
rewrite wH_oldE /wH_old.
rewrite num_occ.num_occ_alt.
apply/eqP/cards1P.
exists i.
apply/setP => n0.
rewrite in_set1 in_set tnth_mktuple 3!mxE.
by case (_ == i).
Qed.

Lemma wH_m_card m0 n0 : #|[set a in 'rV['F_2]_m0 | wH a == n0]| = 'C(m0, n0).
Proof.
rewrite -[in X in _ = X](card_ord m0) -card_draws -2!sum1dep_card.
pose h' := fun s : {set 'I_m0} => \row_(j < m0) (if j \in s then (1%R : 'F_2) else 0%R).
have h'h (i : 'rV_m0) : h' [set i0 | i ord0 i0 == 1%R] == i.
  apply/eqP/rowP => y.
  rewrite !mxE inE.
  case: ifP => [/eqP -> // |].
  move/negbT.
  by rewrite -F2_eq0 => /eqP.
rewrite (reindex_onto (fun x : 'rV_m0 => [set i | x ord0 i == 1%R]) h') /=.
- apply eq_bigl => i.
  rewrite wH_oldE /wH_old num_occ.num_occ_alt h'h andbC /=.
  congr (_ == _).
  apply eq_card => t.
  by rewrite !inE tnth_mktuple.
- move=> s Hs.
  apply/setP => /= i.
  rewrite !inE /h' mxE.
  by case: ifP.
Qed.

Require Import Reals Reals_ext Rbigop Rssr.

Local Open Scope R_scope.
Local Open Scope ring_scope.

Lemma hamming_01 m p :
  \rsum_(u in 'rV['F_2]_m| u \in [set v |(1 >= wH v)%nat])
           ((1 - p) ^ (m - wH u) * p ^ wH u)%R =
  ((1 - p) ^ m + INR m * p * (1 - p) ^ (m - 1))%R.
Proof.
rewrite (bigID [pred i | wH i == O]) /=.
rewrite (big_pred1 (0 : 'rV_m)) /=; last first.
  move=> i /=.
  by rewrite !inE -wH_eq0 andb_idl // => /eqP ->.
rewrite wH0 pow_O subn0 mulR1; f_equal.
transitivity (\rsum_(i | wH (i : 'rV['F_2]_m) == 1%nat) ((1 - p) ^ (m - 1) * p ^ 1)%R).
  transitivity (\rsum_(i|(wH (i : 'rV['F_2]_m) == 1)%nat)
      ((1 - p) ^ (m - wH i) * p ^ wH i)%R).
    apply eq_bigl => /= i.
    rewrite !inE.
    case wH_1 : (wH i == 1)%nat.
      move/eqP in wH_1.
      by rewrite wH_1.
    case wH_0 : (wH i) => [|n1].
      by rewrite eqxx.
    rewrite /= andbT.
    case n1_0 : n1 => [|n2].
      by rewrite wH_0 n1_0 in wH_1.
    by [].
  transitivity (\rsum_(i|wH (i : 'rV['F_2]_m) == 1%N) ((1 - p) ^ (m - 1) * p ^ 1)%R).
    by apply eq_bigr => i /= /eqP ->.
  done.
by rewrite big_const iter_Rplus pow_1 /= -(mulRC p) mulRA -cardsE wH_m_card bin1.
Qed.

Lemma binomial_theorem m p :
 (\rsum_(b|b \in [set: 'rV['F_2]_m]) (1 - p) ^ (m - wH b) * p ^ wH b = 1)%R.
Proof.
transitivity (((1 - p) + p) ^ m); last first.
  rewrite addRC (_ : (p + (1 - p) = 1)%R); last by field.
  by rewrite pow1.
rewrite RPascal.
transitivity (\rsum_(b : 'rV['F_2]_m) ((1 - p) ^ (m - wH b) * p ^ wH b)%R).
  apply eq_bigl => /= i; by rewrite inE.
rewrite (classify_big (fun s : 'rV_m => Ordinal (max_wH' s)) (fun x => ((1 - p) ^ (m - x) * p ^ x))%R) /=.
apply eq_bigr => i _.
do 2 f_equal.
rewrite -wH_m_card.
apply eq_card => /= x; by rewrite !inE.
Qed.

Local Open Scope tuple_ext_scope.

Lemma dH_dH_bitseq n (a b : 'rV_n) :
  dH a b = HammingMetricBitstring.dH_b (tval (rowF2_tuplebool a)) (tval (rowF2_tuplebool b)).
Proof.
rewrite /dH wH_oldE /wH_old.
rewrite /HammingMetricBitstring.dH_b /HammingMetricBitstring.wH_b.
transitivity (N(true | map bool_of_F2 (tuple_of_row (a - b)))).
  rewrite num_occ_sum_bool big_map num_occ_sum.
  apply eq_bigr; by case/F2P.
congr (N(true | _)).
apply/(@eq_from_nth _ true) => [|i Hi].
  by rewrite size_map /addb_seq size_map size_zip !size_tuple minnn.
rewrite size_map size_tuple in Hi.
rewrite (nth_map (0 : 'F_2)); last by rewrite size_tuple.
rewrite /addb_seq.
rewrite (nth_map (true, true)); last by rewrite size_zip 2!size_tuple minnn.
rewrite nth_zip /=; last by rewrite 2!size_map 2!size_tuple.
rewrite /bool_of_F2.
rewrite (_ : _ `_ i = [tuple (a - b) ord0 x | x < n] \_ (Ordinal Hi)); last first.
  rewrite tnth_mktuple.
  rewrite (nth_map (Ordinal Hi)) //; last by rewrite size_enum_ord.
  congr (_ ord0 _).
  by rewrite -[RHS](@nth_ord_enum n (Ordinal Hi)).
rewrite tnth_mktuple !mxE.
rewrite (nth_map (0 : 'F_2)); last by rewrite size_map size_enum_ord.
rewrite (nth_map (0 : 'F_2)); last by rewrite size_map size_enum_ord.
rewrite (_ : _ `_ i = (tuple_of_row a) \_ (Ordinal Hi)); last first.
  rewrite tnth_mktuple (nth_map (Ordinal Hi)); last by rewrite size_enum_ord.
  congr (_ ord0 _).
  apply val_inj => /=.
  by rewrite nth_enum_ord.
rewrite (_ : _ `_ i = (tuple_of_row b) \_ (Ordinal Hi)); last first.
  rewrite tnth_mktuple (nth_map (Ordinal Hi)); last by rewrite size_enum_ord.
  congr (_ ord0 _).
  apply val_inj => /=.
  by rewrite nth_enum_ord.
by rewrite 2!tnth_mktuple oppr_char2 // -bool_of_F2_add_xor.
Qed.

(** The Hamming weight is preserved by permutation (NB: ${\small perm\_mx\,\left( \begin{array}{c} 1 \mapsto 2 \\ 2 \mapsto 1 \end{array} \right) = \left[ \begin{array}{cc} 0 & 1 \\ 1 & 0 \\ \end{array} \right]}$): *)

Lemma perm_on_Sn n (s : 'S_n) : perm_on [set x | x \in enum 'I_n] s.
Proof. apply/subsetP=> /= x _; by rewrite !in_set mem_enum. Qed.

Lemma perm_eq_enum n (s : 'S_n) : perm_eq (enum 'I_n) (map (s^-1)%g (enum 'I_n)).
Proof.
apply uniq_perm_eq.
- by apply enum_uniq.
- rewrite map_inj_uniq; by [apply enum_uniq | apply: perm_inj].
- move=> /= xi.
  case Hi : (xi \in enum 'I_n).
  + symmetry; apply/mapP; exists (s xi).
    * move: (perm_closed xi (perm_on_Sn s)).
      by rewrite !in_set => ->.
    * by rewrite permK.
  + symmetry; apply/mapP; case=> x Hx Hxxi.
    move: (perm_closed x (perm_on_Sn (s^-1)%g)).
    by rewrite !in_set -Hxxi Hx Hi.
Qed.

Lemma wH_perm_mx n (s : 'S_n) (z : 'rV['F_2]_n) : wH (z *m perm_mx s) = wH z.
Proof.
rewrite !wH_oldE /wH_old.
suff -> : tuple_of_row (z *m perm_mx s) = perm_tuple (s^-1)%g (tuple_of_row z).
  by apply: num_occ_perm.
apply eq_from_tnth => n0; by rewrite 3!tnth_mktuple vec_perm_mx !mxE.
Qed.

(** We provide some basic, characteristing properties about row
vectors of Hamming weight 1 %and~2%#and 2# (to be used in %Sect.~\ref{sec:hamming_codes}%#for Hamming codes#): *)

Local Notation "l `b_ i" := (@nth _ false l i) (at level 3, i at level 2).

Lemma wHb_1 : forall n l, size l = n ->
  HammingMetricBitstring.wH_b l = 1%nat ->
  exists i, (i < n /\ l `b_ i = true)%nat /\
    forall j, (j < n -> j <> i -> l `b_ j = false)%nat.
Proof.
elim.
- by case.
- move=> n IH [|[] t] // [Hsz] /=.
  + case=> H.
    exists O.
    split; first by done.
    move=> j Hj Hj'.
    destruct j => //=.
    move: (has_count (pred1 true) t).
    rewrite /HammingMetricBitstring.wH_b /num_occ in H.
    rewrite H ltnn.
    move/negbT/hasPn => K.
    move: (@mem_nth _ false t j).
    rewrite ltnS in Hj.
    rewrite Hsz.
    move/(_ Hj).
    move/K => /= {K}K.
    apply Bool.not_true_is_false.
    by apply/eqP.
  + rewrite add0n.
    case/(IH _ Hsz) => i [[H11 H12] H2].
    exists i.+1.
    split; first by done.
    move=> j Hj Hj'.
    destruct j => //=.
    apply H2 => //.
    contradict Hj'; by subst j.
Qed.

Lemma wH_1 n (x : 'rV['F_2]_n) : wH x = 1%nat ->
  exists i : 'I_n, x 0 i = 1 /\ (forall j : 'I_n, i <> j -> x 0 j = 0).
Proof.
rewrite wH_oldE wH_old_wH_b.
move/wHb_1.
move/(_ n).
rewrite size_tuple.
case/(_ erefl) => i [] [] H1 H2 H3.
exists (Ordinal H1); split.
  transitivity (F2_of_bool true) => //.
  rewrite -H2 /rowF2_tuplebool (nth_map 0); last by rewrite size_tuple.
  rewrite (_ : _ `_ i = (tuple_of_row x) \_ (Ordinal H1)); last first.
    apply set_nth_default; by rewrite size_tuple.
  rewrite tnth_mktuple {1}(F2_0_1 (x 0 (Ordinal H1))); by case: (x 0 _ != 0).
move=> j Hj.
transitivity (F2_of_bool false) => //.
rewrite -(H3 j) //; last first.
  move=> ?; subst i.
  by apply/Hj/val_inj.
rewrite /rowF2_tuplebool (nth_map 0); last by rewrite size_tuple.
rewrite (_ : _ `_ j = (tuple_of_row x) \_ j); last first.
    apply set_nth_default; by rewrite size_tuple.
rewrite tnth_mktuple {1}(F2_0_1 (x 0 j)); by case: (x 0 _ != 0).
Qed.

Lemma wHb_2 : forall n l, size l = n ->
  HammingMetricBitstring.wH_b l = 2%nat ->
  (exists i, exists k, (i < n /\ l `b_ i = true) /\
    (k < n /\ l `b_ k = true) /\ i <> k /\
    forall j, j < n -> j <> i -> j <> k -> l `b_ j = false)%nat.
Proof.
elim.
- by case.
- case.
  move=> IH l Hsz Hcount.
  rewrite /HammingMetricBitstring.wH_b /num_occ in Hcount.
  move: (count_size (pred1 true) l).
  by rewrite Hcount Hsz ltnn.
- move=> n IH [|[] t] // [Hsz] /=.
  + rewrite add1n.
    case=> X.
    exists O.
    case: (@wHb_1 _ _ Hsz X) => k [[Hk1 Hk11] Hk2].
    exists k.+1.
    split; first by done.
    split.
    split; last by done.
    by apply leq_ltn_trans with n.+1.
    split; first by done.
    move=> j Hj Hj' Hjk.
    destruct j => //=.
    apply Hk2 => //.
    contradict Hjk; by rewrite Hjk.
  + rewrite add0n.
    case/(IH _ Hsz) => i [k [[H11 H12] [H21 [H221 H222]]]].
    exists i.+1, k.+1.
    split; first by done.
    split; first by done.
    split.
    contradict H221; by case: H221.
    move=> j Hj Hj' Hjk.
    destruct j => //=.
    apply H222 => //.
    contradict Hj'; by subst j.
    contradict Hjk; by subst j.
Qed.

Lemma wH_2 n (x : 'rV['F_2]_n) : wH x = 2%nat ->
  exists i j : 'I_n, i <> j /\ x 0 i = 1 /\ x 0 j = 1 /\
  (forall k : 'I_n, i <> k -> j <> k -> x 0 k = 0).
Proof.
rewrite wH_oldE wH_old_wH_b.
move/wHb_2.
move/(_ n).
rewrite size_tuple.
case/(_ erefl) => i [] k [] [] H1 [] H2 [] [] H3 [] H4 [] H5 H6.
exists (Ordinal H1), (Ordinal H3); split.
  by case.
split.
  transitivity (F2_of_bool true) => //.
  rewrite -H2 /rowF2_tuplebool (nth_map 0); last by rewrite size_tuple.
  rewrite (_ : _ `_ i = (tuple_of_row x) \_ (Ordinal H1)); last first.
    apply set_nth_default; by rewrite size_tuple.
  rewrite tnth_mktuple {1}(F2_0_1 (x 0 (Ordinal H1))); by case: (x 0 _ != 0).
split.
  transitivity (F2_of_bool true) => //.
  rewrite -H4 /rowF2_tuplebool (nth_map 0); last by rewrite size_tuple.
  rewrite (_ : _ `_ k = (tuple_of_row x) \_ (Ordinal H3)); last first.
    apply set_nth_default; by rewrite size_tuple.
  rewrite tnth_mktuple {1}(F2_0_1 (x 0 (Ordinal H3))); by case: (x 0 _ != 0).
move=> j ij kj.
transitivity (F2_of_bool false) => //.
rewrite -(H6 j) //; last first.
- move=> ?; subst k.
  by apply/kj/val_inj.
- move=> ?; subst i.
  by apply/ij/val_inj.
rewrite /rowF2_tuplebool (nth_map 0); last by rewrite size_tuple.
rewrite (_ : _ `_ j = (tuple_of_row x) \_ j); last first.
    apply set_nth_default; by rewrite size_tuple.
rewrite tnth_mktuple {1}(F2_0_1 (x 0 j)); by case: (x 0 _ != 0).
Qed.

Lemma card_dH {n} (t t' : n.-tuple 'F_2) :
  #| [pred i | t' \_ i != t \_ i ] | = dH (row_of_tuple t) (row_of_tuple t').
Proof.
rewrite /dH wH_oldE /wH_old num_occ.num_occ_alt /=.
apply eq_card => /= i.
rewrite inE.
move H : (_ \in _) => [|].
  symmetry.
  rewrite F2_eq1.
  move: H; apply contra.
  by rewrite tnth_mktuple !mxE subr_eq0 eq_sym.
move/negbT in H.
rewrite negbK in H.
apply/esym/negbTE.
rewrite -F2_eq0 tnth_mktuple !mxE.
move/eqP : H => ->.
by rewrite subr_eq0.
Qed.

Local Open Scope vec_ext_scope.

(* TODO: rename? move? *)
Lemma card_dH_vec {n} (t t' : 'rV['F_2]_n) :
  #| [pred i | t' ``_ i != t ``_ i ] | = dH t t'.
Proof.
rewrite /dH wH_oldE /wH_old num_occ.num_occ_alt /=.
apply eq_card => /= i.
rewrite inE.
move H : (_ \in _) => [|].
  symmetry.
  rewrite F2_eq1.
  apply: contra H.
  by rewrite tnth_mktuple !mxE subr_eq0 eq_sym.
move/negbT in H.
rewrite negbK in H.
apply/esym/negbTE.
rewrite -F2_eq0 tnth_mktuple !mxE.
move/eqP : H => ->.
by rewrite subr_eq0.
Qed.

Lemma card_dHC {n} (t t': 'rV['F_2]_n) :
  #| [pred i | t' ``_ i == t ``_ i ] | = (n - dH t t')%nat.
Proof.
move: (cardC [pred i | t' ``_i == t ``_i ]).
rewrite card_ord => H.
by rewrite -[in X in _ = (X - _)%nat]H -card_dH_vec -addnBA // subnn addn0.
Qed.

Local Close Scope vec_ext_scope.

Local Close Scope tuple_ext_scope.

End HammingMetric_F_2.

(** * Encoding of Naturals as Vectors *)

(** Here is a function that transforms natural numbers into their binary encodings
   (presented in the form of column-vectors) (e.g., $nat2bin\,3\,1 = \left[ \begin{array}{c} 0 \\ 0 \\ 1 \end{array} \right]$, $nat2bin\,3\,2 = \left[ \begin{array}{c} 0 \\ 1 \\ 0 \end{array} \right]$, $nat2bin\,3\,3 = \left[ \begin{array}{c} 0 \\ 1 \\ 1 \end{array} \right]$, etc.): *)

Definition rV_of_nat (n i : nat) : 'rV_n := bitseq_F2row (size_nat2bin i n).

Definition cV_of_nat (n i : nat) : 'cV_n := (rV_of_nat n i)^T.

Lemma rV_of_nat_neq0 n i : i <> O -> i < expn 2 n -> rV_of_nat n i <> 0.
Proof.
move=> Hi W.
rewrite /rV_of_nat /bitseq_F2row /bitseq_to_row.
move: (nat2bin_nseq_false i n Hi W) => H.
contradict H.
apply eq_from_nth with false.
- by rewrite /= size_pad_seqL size_nseq.
- move=> j Hj.
  rewrite (_ : size _ = n) in Hj; last first.
    (* TODO: lemma? *)
    apply/eqP. by apply size_nat2bin.
  move/rowP : H => /(_ (Ordinal Hj)).
  rewrite !mxE /= => Hj'.
  rewrite nth_nseq Hj.
  by apply F2_of_bool_0_inv.
Qed.

Lemma rV_of_nat_inj n i j : nat_of_pos i < expn 2 n -> nat_of_pos j < expn 2 n ->
  rV_of_nat n (nat_of_pos i) = rV_of_nat n (nat_of_pos j) -> i = j.
Proof.
move=> Hi Hj.
rewrite /rV_of_nat /bitseq_F2row /bitseq_to_row.
move/row_nth => X.
have Htmp : size (nat2bin (nat_of_pos i) n) <= n.
  move: (size_nat2bin (nat_of_pos i) n).
  by move/eqP => ->.
apply X in Htmp => //; last first.
  apply (@trans_eq _ _ n).
  - by apply/eqP/size_nat2bin.
  - by apply/esym/eqP/size_nat2bin.
rewrite /nat2bin in Htmp.
move: (N2bitseq_leading_bit (bin_of_nat (nat_of_pos i))) => U.
lapply U; last by apply bin_of_nat_nat_of_pos_not_0.
case=> ik Hik.
rewrite Hik in Htmp.
move: (N2bitseq_leading_bit (bin_of_nat (nat_of_pos j))) => V.
lapply V; last by apply bin_of_nat_nat_of_pos_not_0.
case=> jk Hjk.
rewrite Hjk in Htmp.
destruct n.
- rewrite expn0 in Hi.
  move: (@nat_of_pos_not_0 i) => W.
  by destruct (nat_of_pos i).
- apply pad_seqL_leading_true_inj in Htmp; last 2 first.
    have XX : size (true :: ik) <= n.+1.
      rewrite -Hik size_rev.
      apply size_N2bitseq_ub => //.
      by apply nat_of_pos_not_0.
    by rewrite /= ltnS in XX.
  have XX : size (true :: jk) <= n.+1.
    rewrite -Hjk size_rev.
    apply size_N2bitseq_ub => //.
    by apply nat_of_pos_not_0.
    by rewrite /= ltnS in XX.
  subst ik.
  rewrite -Hjk in Hik.
  move/rev_inj : Hik.
  by move/N2bitseq_inj/bin_of_nat_inj/nat_of_pos_inj.
Qed.

Lemma rV_of_natD_neq0 r i j : i <> j ->
  i <> O -> j <> O -> i < expn 2 r -> j < expn 2 r -> rV_of_nat r i + rV_of_nat r j <> 0.
Proof.
move=> Hij Hi Hj Hin Hjn.
destruct i => //.
destruct j => //.
contradict Hij.
apply F2_addmx0 in Hij.
have [ii Hii] : exists ii, i.+1 = nat_of_pos ii.
  exists (BinPos.P_of_succ_nat i).
  rewrite -Pnat.nat_of_P_o_P_of_succ_nat_eq_succ.
  by rewrite BinPos_nat_of_P_nat_of_pos.
have [jj Hjj] : exists jj, j.+1 = nat_of_pos jj.
  exists (BinPos.P_of_succ_nat j).
  rewrite -Pnat.nat_of_P_o_P_of_succ_nat_eq_succ.
  by rewrite BinPos_nat_of_P_nat_of_pos.
rewrite Hii Hjj in Hij.
apply rV_of_nat_inj in Hij; last 2 first.
  by rewrite -Hii.
  by rewrite -Hjj.
rewrite Hjj; by subst ii.
Qed.

Lemma wH_two_pow p m : p < m -> wH (rV_of_nat m (expn 2 p)) = 1%nat.
Proof.
move=> p_m.
rewrite wH_oldE wH_old_wH_b /rV_of_nat /bitseq_F2row /bitseq_to_row /rowF2_tuplebool /=.
rewrite /HammingMetricBitstring.wH_b /num_occ.num_occ count_map.
rewrite (eq_count (a2 := pred1 1)); last by move=> ? /=; rewrite eqb_id F2_eq1.
rewrite -sum1_count big_map /=.
rewrite (eq_bigl (pred1 (Ordinal (rev_ord_proof (Ordinal p_m))))); last first.
  move=> i /=; symmetry; rewrite -[in X in X = _]val_eqE /=.
  rewrite mxE nat2bin_two_pow // nth_cat size_nseq.
  case: ifP => [ i_m | /negbT ].
    rewrite nth_nseq i_m /=.
    apply/negbTE; by rewrite neq_ltn i_m.
  rewrite -leqNgt leq_eqVlt; case/orP => [/eqP Heq | m_i ].
    by rewrite -Heq subnn /= !eqxx.
  rewrite -[in X in _ = X]subnSK //= nth_nseq.
  transitivity false; last by case: ifP.
  apply/negbTE; by rewrite neq_ltn m_i orbC.
by rewrite /= big_filter_cond /= big_pred1_eq.
Qed.

Lemma wH_3 : forall n, 3 <= n -> wH (rV_of_nat n 3) = 2.
Proof.
case => // [] // [] // [] // n _.
rewrite /wH /num_occ.num_occ /rV_of_nat.
set tmp := tuple_of_row _.
have -> : tmp = [tuple of rev (1 :: 1 :: nseq n.+1 (0 : 'F_2))].
  apply eq_from_tnth => i.
  rewrite tnth_mktuple /bitseq_F2row /bitseq_to_row /nat2bin mxE.
  rewrite (tnth_nth 0).
  rewrite -(nth_map _ 0); last by rewrite size_cat size_nseq /= addn2.
  congr (_ `_ i).
  rewrite /= map_cat /= -!cat_rcons cats0 (_ : 0 :: nseq n 0 = nseq n.+1 0) //.
  by rewrite 2!rev_cons nseq_S -cat_rcons cats0 rev_rcons /= rev_nseq map_nseq.
rewrite /=.
set y := _ :: _.
rewrite (_ : y = [:: 1; 1; 0] ++ nseq n 0) // rev_cat count_cat /= -(@eq_in_count _ pred0).
  by rewrite count_pred0.
move=> a /=.
rewrite rev_nseq.
by move/mem_nseq/eqP => ->.
Qed.

Lemma wH_7 : forall n, 3 <= n -> wH (rV_of_nat n 7) = 3.
Proof.
case => // [] // [] // [] // n _.
rewrite /wH /num_occ.num_occ /rV_of_nat.
set x := tuple_of_row _.
have -> : x =[tuple of rev (1 :: 1 :: 1 :: nseq n (0 : 'F_2))].
  rewrite /x {x}.
  apply eq_from_tnth => i.
  rewrite tnth_mktuple /bitseq_F2row /bitseq_to_row /nat2bin mxE (tnth_nth 0).
  rewrite -(nth_map _ 0); last first.
    rewrite /pad_seqL /=.
    destruct n => //=.
    by rewrite size_cat size_nseq /= addn3.
  congr (_ `_ i).
  rewrite /= !rev_cons /= /pad_seqL /= -addn3 addnK map_cat /= map_nseq /=.
  by rewrite rev_nseq -!cats1 /= -!catA.
rewrite /=.
set y := _ :: _.
rewrite (_ : y = [:: 1; 1; 1] ++ nseq n 0) // rev_cat count_cat /=.
rewrite -(@eq_in_count _ pred0).
  by rewrite count_pred0.
move=> a /=.
rewrite rev_nseq.
by move/mem_nseq/eqP => ->.
Qed.

Lemma rev7_bin : forall m, 2 < m ->
  rev (N2bitseq (BinNums.Npos (xOs (m - 3) (BinNums.xI 3)))) =
  true :: true :: true :: nseq (m - 3) false.
Proof.
elim=> //.
case=> //.
case=> //.
case=> //.
move=> m.
move/(_ Logic.eq_refl).
rewrite -{1}addn3 -addnBA // subnn addn0 => IH _.
rewrite -{1}addn4 -addnBA //= (_ : 4 - 3 = 1)%nat // addn1 /=.
rewrite rev_cons IH /=.
rewrite -{1}addn3 -addnBA // subnn addn0.
rewrite -cats1.
by rewrite -nseq_S.
Qed.

Lemma rev_nat2bin_7 n : 2 < n ->
  rev (nat2bin 7 n) = nat2bin (7 * expn 2 (n - 3)) n.
Proof.
move=>Hn.
rewrite /nat2bin.
rewrite (bin_of_nat_rev7 _ Hn).
rewrite (@bin_of_nat_7 n) //=.
rewrite (rev7_bin Hn) /=.
rewrite {1}/pad_seqL /=.
rewrite Hn.
rewrite rev_cat /=.
rewrite /pad_seqL /=.
rewrite size_nseq.
case: ifP => H //.
  rewrite (_ : n - (n - 3).+3 = 0)%nat //; last first.
  destruct n => //.
  destruct n => //.
  destruct n => //.
  rewrite -addn3.
  by rewrite -(@addnBA n 3 3) // subnn addn0 addn3 subnn.
by rewrite rev_nseq.
suff : False by done.
move/negbT in H.
rewrite -leqNgt in H.
destruct n => //.
destruct n => //.
destruct n => //.
by rewrite -{2}addn3 -addnBA // subnn addn0 ltnn in H.
Qed.

Lemma wH_7_rev7 n : 2 < n ->
  wH (rV_of_nat n 7) = wH (rV_of_nat n (7 * expn 2 (n - 3))).
Proof.
move=> Hn.
rewrite 2!wH_oldE /wH_old /rV_of_nat.
set lhs := tuple_of_row _.
set rhs := tuple_of_row _.
suff : rev lhs = rhs.
  move=> <-.
  by apply num_occ.num_occ_rev.
rewrite /lhs /rhs.
rewrite /bitseq_F2row.
rewrite -!bitseq_to_rowK.
rewrite -map_rev.
f_equal.
by rewrite rev_nat2bin_7.
Qed.

Definition nat_of_rV n (y : 'rV_n) :=
  BinNat.N.to_nat (bitseq2N (map bool_of_F2 (tuple_of_row y))).

Lemma nat_of_rV_up n (y : 'rV_n) : nat_of_rV y < expn 2 n.
Proof. by rewrite /nat_of_rV bitseq2N_up // size_map size_tuple. Qed.

Lemma nat_of_rV_0 k : nat_of_rV (0 : 'rV_k) = O.
Proof.
rewrite /nat_of_rV.
set tmp := map _ _.
have -> : tmp = nseq k false.
  rewrite {}/tmp  /= (_ : false = bool_of_F2 0) // -map_nseq.
  congr ([seq _ | i <- _]).
  apply eq_from_nth with 0.
  - by rewrite size_map size_enum_ord size_nseq.
  - move=> i Hi.
    rewrite size_map size_enum_ord in Hi.
    rewrite nth_nseq Hi (nth_map (Ordinal Hi)); last by rewrite size_enum_ord.
    by rewrite mxE.
by rewrite bitseq2N_nseq_false.
Qed.

Require Import BinNat.

Lemma nat_of_rV_eq0 n (y : 'rV_n) : (nat_of_rV y == O) = (y == 0).
Proof.
case Hlhs : (nat_of_rV _ == O); last first.
  symmetry; apply/negP => /eqP abs; subst y.
  by rewrite nat_of_rV_0 eqxx in Hlhs.
symmetry; apply/eqP.
move/eqP : Hlhs.
rewrite /nat_of_rV [X in _ = X -> _](_ : O = N.to_nat 0) //.
move/Nnat.N2Nat.inj/bitseq2N_0.
rewrite size_map size_tuple.
move/(_ _ erefl) => Htmp.
apply tuple_of_row_inj, val_inj.
rewrite -(map_id (val (tuple_of_row y))).
transitivity (map (F2_of_bool \o bool_of_F2) (val (tuple_of_row y))).
  apply eq_in_map => i /= _; by rewrite bool_of_F2K.
by rewrite map_comp Htmp row_to_seq_0 (_ : 0 = F2_of_bool false) // map_nseq.
Qed.

Definition nat_of_cV n (y : 'cV_n) : nat := nat_of_rV y^T.

Lemma nat_of_cV_0 k : nat_of_cV (0 : 'cV_k) = O.
Proof. by rewrite /nat_of_cV trmx0 nat_of_rV_0. Qed.

Lemma nat_of_rV_tr n (y : 'rV_n) : nat_of_rV y = nat_of_cV (y^T).
Proof. by rewrite /nat_of_cV trmxK. Qed.

Lemma mulmx_rV_of_nat_row n m (M : 'M_(m, n)) (k : 'I_m) :
  (rV_of_nat m (expn 2 k)) *m M = row (Ordinal (rev_ord_proof k)) M.
Proof.
rewrite /rV_of_nat.
apply/rowP => x.
rewrite !mxE /=.
transitivity (\sum_(j < m) (F2_of_bool (nth false (nat2bin (expn 2 k) m) j) * M j x)).
  apply eq_bigr => i _; by rewrite mxE.
rewrite nat2bin_two_pow //.
pose mk := Ordinal (rev_ord_proof k).
rewrite -/mk (bigID (pred1 mk)) /= big_pred1_eq.
set tmp1 := nth _ _ _.
have -> : tmp1 = true.
  by rewrite /tmp1 {tmp1} nth_cat size_nseq {1}/mk ltnn subnn.
rewrite mul1r.
set lhs := \sum_ (_ | _) _.
suff -> : lhs = 0 by rewrite addr0.
transitivity (\sum_(i | i != mk) (0 : 'F_2)).
  apply eq_bigr => i imk.
  set rhs := nth _ _ _.
  suff -> : rhs = false by rewrite mul0r.
  rewrite /rhs nth_cat size_nseq.
  case: ifP => Hcase; first by rewrite nth_nseq Hcase.
  rewrite (_ : true :: _ = [:: true] ++ nseq k false) // nth_cat /=.
  case: ifP => Hcase2.
    suff : False by done.
    clear -Hcase2 imk Hcase.
    have {Hcase2}Hcase2 : i <= m - k.+1 by rewrite ltnS leqn0 in Hcase2.
    rewrite leq_eqVlt Hcase orbC /= in Hcase2.
    move: Hcase2; by apply/negP.
  rewrite nth_nseq; by case: ifP.
by rewrite big_const iter_addr0.
Qed.

Lemma tuple_of_row_ord0 (s : 'rV['F_2]_0) : tuple_of_row s = [tuple of [::]].
Proof. apply eq_from_tnth; by case. Qed.

Lemma nat_of_rV_ord0 (s : 'rV_0) : nat_of_rV s = O.
Proof. by rewrite /nat_of_rV N_bin_to_nat tuple_of_row_ord0. Qed.

Lemma rV_of_nat_0 n : rV_of_nat n 0 = 0.
Proof.
rewrite /rV_of_nat /bitseq_F2row /bitseq_to_row nat2bin_0_n.
apply/rowP => b /=.
rewrite 2!mxE nth_nseq; by case: ifP.
Qed.

Lemma nat_of_rVK n (s : 'rV_n) : rV_of_nat n (nat_of_rV s) = s.
Proof.
destruct n as [|n].
  apply/rowP; by case.
apply/rowP => y.
rewrite mxE /nat_of_rV.
set tmp := [seq bool_of_F2 i | i <- _].
rewrite [X in nat2bin _ X](_ : _ = size tmp); last by rewrite size_map size_tuple.
rewrite bitseq2NK; last by rewrite size_tuple.
rewrite /tmp (nth_map (0 : 'F_2)); last by rewrite size_tuple.
rewrite bool_of_F2K (_ : _ `_ _ = tnth (tuple_of_row s) y); last first.
  apply set_nth_default; by rewrite size_tuple.
by rewrite tnth_mktuple.
Qed.
